@node_all = (-1, -1, -1, -1);
@node_me  = (2, 550, 180, 0);
# --------------------------------------------------------------------
# Parse 5D-addr in form "z:n/f.p@domain" into array
sub s2addr
{
  my ($z, $n, $f, $p, $domain);
  if (@_ != 1 && @_ != 5) { return undef; }
  elsif (@_ == 1) { ($z, $n, $f, $p) = @node_me; }
  else { ($z, $n, $f, $p) = ($_[1], $_[2], $_[3], $_[4]); }
  $p = 0;
  $_ = $_[0];
  # handle zone, if any
  #print "pre-zone: [$_]\n";
  if (m/^\*@/) { $_ = "*:*/*.$_" }
    elsif (m/^\*\./) { $_ = "*:*/$_" }
    elsif (m/^\*$/) { $_ = "*:*/*.$_"; }
  if (m/^(.+):/) {
    my $x = $1;
    if ( $x eq '*' ) { $z = -1; }
    elsif ( $x =~ m/^\d+$/ ) { $z = $x; }
    else { return undef; }
    s/^.+://;
  }
  # handle net, if any
  #print "pre-net: [$_]\n";
  if (m/^\*@/) { $_ = "*/*.$_" }
    elsif (m/^\*\./) { $_ = "*/$_" }
    elsif (m/^\*$/) { $_ = "*/*.$_" }
  if (m!^(.+)/!) {
    my $x = $1;
    if ( $x eq '*' ) { $n = -1; }
    elsif ( $x =~ m/^\d+$/ ) { $n = $x; }
    else { return undef; }
    s!^.+/!!;
  }
  # handle node
  #print "pre-node: [$_]\n";
  if (m!^/?(?:\*(@.*)|\*$)!) { $_ = "*.*$1"; }
  if (!m/^[.@]/) {
    if (m!^/?([^.@]+)!) {
      my $x = $1;
      if ( $x eq '*' ) { $f = -1; }
      elsif ( $x =~ m/^\d+$/ ) { $f = $x; }
      else { return undef; }
      s!^[^@.]+([@.])?!\1!;
    }
  }
  # handle point, if any
  #print "pre-point: [$_]\n";
  if (m!^\.([^@]+)!) { 
    my $x = $1;
    if ( $x eq '*' ) { $p = -1; } 
    elsif ( $x =~ m/^\d+$/ ) { $p = $x; }
    else { return undef; }
    s!^\.([^@]+)!!;
  }
  # handle domain, if any
  if (m!^@(.+)!) {
    $domain = $1;
  }
  #print "return: $z:$n/$f.$p\n";
  my @res = ($z, $n, $f);
  if ($p || ($domain ne '')) { push @res, $p; }
  if ( $domain ne '' ) { push @res, $domain; }
  return @res;
}
# --------------------------------------------------------------------
# Convert address in array to string
sub addr2s
{
  if ( @_ < 3 ) { return undef; }
  my ($z, $n, $f, $p, $domain) = @_;
  $z = '*' if $z == -1;
  $n = '*' if $n == -1;
  $f = '*' if $f == -1;
  $p = '*' if $p == -1;
  my $s = "$z:$n/$f";
  $s .= ".$p" if (@_ >= 4 && $p > 0);
  $s .= "\@$domain" if (@_ >=5 && $domain ne '');
  return $s;
  #if (@_ == 5) { return "$z:$n/$f.$p\@$domain"; }
  #elsif (@_ == 4) { return "$z:$n/$f.$p"; }
  #else { return "$z:$n/$f"; }
}
# --------------------------------------------------------------------
# Convert address in array to string - unique to z:n/f.p
sub addr4s
{
  if ( @_ < 3 ) { return undef; }
  my ($z, $n, $f, $p) = @_;
  $z = '*' if $z == -1;
  $n = '*' if $n == -1;
  $f = '*' if $f == -1;
  $p = '*' if $p == -1;
  if (($p ne '*') && ($p == 0)) { return "$z:$n/$f"; } 
  else { return "$z:$n/$f.$p"; }
}
# --------------------------------------------------------------------
# Parse fqfa-addr in form "domain#z:n/f.p" into array
sub fqfa2addr
{
  if ( $_[0] =~ m!^(.*?)#(\d+:\d+/\d+\.\d+)$! ) { return s2addr($2.'@'.$1); }
  else { return s2addr($_[0]); }
}
# --------------------------------------------------------------------
# Compare two fido addresses
sub acmp
{
  my @a1 = @{$_[0]};
  my @a2 = @{$_[1]};
  for (my $i = 0; $i < 4; $i++) {
    if ( ($a1[$i] != -1) && ($a2[$i] != -1) && ($a1[$i] != $a2[$i]) ) { return 0; }
  }
  if (@a1 < 5 && @a2 < 5) { return 1; }
  # compare domains
  if ($a1[4] eq '*' || $a2[4] eq '*') { return 1; }
  my ($dom1, $dom2);
  eval {
    use misc;
    $dom1 = (@a1 < 5) ? ::domain(@a1) : $a1[4];
    $dom2 = (@a2 < 5) ? ::domain(@a2) : $a2[4];
  };
  return int(lc($dom1) eq lc($dom2));
}

1;
