use fidoaddr;
use msg;
use misc;
use config;

%MODE = ('>'=>'read-only', '<'=>'write-only', '+'=>'read-write');

$Mgr::unlisted  = "Your are not allowed to make AreaMgr requests\n\n---\n";
$Mgr::wrong_pwd = "Your AreaMgr password is incorrect\n\n---\n";

# --------------------------------------------------------------------
# do_link($mask, \%msg)
#    return reply message, undef if nothing is done
sub do_link
{
  my $mask = $_[0]; my %msg = %{$_[1]};
  my $multi = $mask =~ /[*?]/;
  my $reply;
  $mask =~ s/([^A-Za-z0-9_?*])/\\\1/g;
  $mask =~ s/[*]/.*?/g;
  $mask =~ s/[?]/./g;
  foreach my $tag (sort keys %areas) {
    if ($tag =~ /^[*?]/) { next; }
    if ($tag !~ /^$mask$/i) { next; }
    say 5, "> connecting: matched $tag";
    my %area = %{$areas{$tag}};
    my $mode = area_node($areas{$tag}, $msg{afrom});
    if (!defined $mode) {
      if (!area_avail($tag, $msg{afrom})) { 
        $reply .= "$tag: permission denied\n" unless $multi; next; 
      }
      modify_link($tag, '+'.$area{restrict}.addr2s(@{$msg{afrom}}));
      $reply .= "$tag: connected $MODE{$area{restrict}}\n"; next;
    } else {
      if ($mode eq $area{restrict}) { 
        $reply .= "$tag: already connected\n" unless $multi; next; 
      }
      modify_link($tag, '+'.$area{restrict}.addr2s(@{$msg{afrom}}));
      $reply .= "$tag: re-connected ".$MODE{substr $area{restrict}.'+', 0, 1}."\n"; next;
    }
  }
  return $reply;
}
# --------------------------------------------------------------------
# do_unlink($mask, \%msg)
#    return reply message, undef if nothing is done
sub do_unlink
{
  my $mask = $_[0]; my %msg = %{$_[1]};
  my $multi = $mask =~ /[*?]/;
  my $reply;
  $mask =~ s/([^A-Za-z0-9_?*])/\\\1/g;
  $mask =~ s/[*]/.*?/g;
  $mask =~ s/[?]/./g;
  foreach my $tag (sort keys %areas) {
    if ($tag =~ /^[*?]/) { next; }
    if ($tag !~ /^$mask$/i) { next; }
    say 5, "> disconnecting: matched $tag";
    my %area = %{$areas{$tag}};
    if (!%area) { $reply .= "$tag: unknown fileecho\n" unless $multi; next; }
    my $mode = area_node($areas{$tag}, $msg{'afrom'});
    if (!defined $mode) { $reply .= "$tag: not connected\n" unless $multi; next; }
    modify_link($tag, '-'.addr2s(@{$msg{'afrom'}}));
    $reply .= "$tag: disconnected\n";
  }
  return $reply;
}
# --------------------------------------------------------------------
# mgr($text, \%msg): process request $text from message %msg
sub mgr
{
  my $subj, $i;
  my %msg = %{$_[1]};
  my @arr = split /\n/, $_[0];
  my $line = 0; my $reply = '';
  # some general stuff
  my %repl = (from=>$ROBOT, afrom=>use_aka($msg{afrom}), 
              to=>$msg{from}, ato=>$msg{afrom},
              subj=>'AreaMgr Reply', 
              flags=>$Msg::PVT | $Msg::LOC | flag_msg( $nodes{ addr4s(@{$msg{afrom}}) }{flag_mail} ) );
  @Tpl::node = @{$msg{afrom}};
  $Tpl::aka  = addr2s( @{use_aka(@Tpl::node)} );
  # process
  foreach (@arr) {
    $line++;
    study $_;
    if (/^\s*$/) { next; }
    elsif ( /^---/ ) { last; }
    # %HELP
    elsif ( /^\s*%h(elp)?/i ) {
      $Tpl::from = $msg{'from'}; $Tpl::sfrom = addr2s(@Tpl::node);
      $repl{'text'} = do 'tpl/help.pl';
      #Msg::fwrite(\%repl, $config{mailout}.'/');
      send_mail(\%repl);
    }
    # %STATUS
    elsif ( /^\s*%s(tatus)?/i ) {
      $Tpl::node = $nodes{ addr4s(@{$msg{afrom}}) };
      $repl{'text'} = do 'tpl/status.pl';
      #Msg::fwrite(\%repl, $config{mailout}.'/');
      send_mail(\%repl);
    }
    # %LIST
    elsif ( /^\s*%l(ist)?/i ) {
      $Tpl::type = 'all';
      $Tpl::empty = '  There are no fileechos available to you';
      @Tpl::areas = ();
      while ( my ($k, $v) = each %areas ) { push @Tpl::areas, $k unless $k =~ /^[?]/; }
      do 'tpl/list.pl';
      $subj = $repl{'subj'};
      for ($i = 0; $i < @Tpl::res; $i++) {
        $repl{'text'} = $Tpl::res[$i];
        $repl{'subj'} = "List of all fileechos";
        $repl{'subj'} .= " [".($i+1)."/".($#Tpl::res+1)."]" if @Tpl::res > 1;
        send_mail(\%repl);
      }
      $repl{'subj'} = $subj;
    }
    # %QUERY
    elsif ( /^\s*%q(uery)?/i ) { 
      $Tpl::type = 'linked';
      $Tpl::empty = '  You are not connected to any fileechos';
      @Tpl::areas = ();
      while ( my ($k, $v) = each %areas ) { push @Tpl::areas, $k unless ($k =~ /^[?]/) || !defined area_node($areas{$k}, $msg{afrom}); }
      do 'tpl/list.pl';
      $subj = $repl{'subj'};
      for ($i = 0; $i < @Tpl::res; $i++) {
        $repl{'text'} = $Tpl::res[$i];
        $repl{'subj'} = "List of connected fileechos";
        $repl{'subj'} .= " [".($i+1)."/".($#Tpl::res+1)."]" if @Tpl::res > 1;
        send_mail(\%repl);
      }
      $repl{'subj'} = $subj;
      #$repl{'text'} = do 'tpl/list.pl';
      #Msg::fwrite(\%repl, $config{mailout}.'/');
      #send_mail(\%repl);
    }
    # %UNLINKED
    elsif ( /^\s*%u(nlinked)?/i ) {
      $Tpl::type = 'not linked';
      $Tpl::empty = '  You are connected to all available fileechos';
      @Tpl::areas = ();
      while ( my ($k, $v) = each %areas ) { push @Tpl::areas, $k unless ($k =~ /^[?]/) || defined area_node($areas{$k}, $msg{afrom}); }
      do 'tpl/list.pl';
      $subj = $repl{'subj'};
      for ($i = 0; $i < @Tpl::res; $i++) {
        $repl{'text'} = $Tpl::res[$i];
        $repl{'subj'} = "List of unconnected fileechos";
        $repl{'subj'} .= " [".($i+1)."/".($#Tpl::res+1)."]" if @Tpl::res > 1;
        send_mail(\%repl);
      }
      $repl{'subj'} = $subj;
      #$repl{'text'} = do 'tpl/list.pl';
      #Msg::fwrite(\%repl, $config{mailout}.'/');
      #send_mail(\%repl);
    }
    # %AVAIL
    elsif ( /^\s*%avail/i ) {
      my @txts = forward_avail($msg{afrom});
      $subj = $repl{'subj'};
      for ($i = 0; $i < @txts; $i++) {
        $repl{'text'} = $txts[$i];
        $repl{'subj'} = "List of requestable fileechos";
        $repl{'subj'} .= " [".($i+1)."/".($#txts+1)."]" if @txts > 1;
        send_mail(\%repl);
      }
      $repl{'subj'} = $subj;
    }
    # %NOTIFY <on|off>
    elsif ( /^\s*%notify\s+(on|off)/i ) {
      modify_node($msg{afrom}, (lc $1 eq 'on' ? '+' : '-').'notify');
      $reply .= "Notify function for your system is now ".(lc $1)."\n";
    }
    # %MESSAGE <on|off>
    elsif ( /^\s*%message\s+(on|off)/i ) {
      modify_node($msg{afrom}, (lc $1 eq 'on' ? '+' : '-').'msg');
      $reply .= "Sent files notification for your system is now ".(lc $1)."\n";
    }
    # %NEWS <on|off>
    elsif ( /^\s*%news\s+(on|off)/i ) {
      modify_node($msg{afrom}, (lc $1 eq 'on' ? '+' : '-').'news');
      $reply .= "Sending news about area auto-creating is now ".(lc $1)."\n";
    }
    # %PAUSE
    elsif ( /^\s*%(pause|passive)/i ) {
      modify_node($msg{afrom}, '+pause');
      $reply .= "Sending files to your system is now paused\n";
    }
    # %RESUME
    elsif ( /^\s*%(resume|active)/i ) {
      modify_node($msg{afrom}, '-pause');
      $reply .= "Sending files to your system is now resumed\n";
    }
    # %+ALL, %+GROUP <group>
    elsif ( /^\s*%[+]all/i || /^\s*%[+]group\s+(\S+)/i ) {
      if ( defined $1 ) { say 5, "> connecting group $1"; } else { say 5, "> connecting all areas"; }
      while ( my ($tag, $area) = each %areas ) {
        if ( defined $1 && ($area->{group} ne $1) ) { next; }
        if ( area_avail($tag, $msg{afrom}) ) {
          my $mode = area_node($area, $msg{afrom});
          if ( !defined $mode ) {
            modify_link($tag, '+'.$area->{restrict}.addr2s(@{$msg{afrom}}));
            $reply .= "$tag: connected $MODE{$area{restrict}}\n"; next;
          } elsif ($mode ne $area->{restrict}) { 
            modify_link($tag, '+'.$area->{restrict}.addr2s(@{$msg{afrom}}));
            $reply .= "$tag: re-connected ".$MODE{substr $area->{restrict}.'+', 0, 1}."\n"; next;
          }
        }
      }
    }
    # %-ALL, %-GROUP <group>
    elsif ( /^\s*%[-]all/i || /^\s*%[-]group\s+(\S+)/i ) {
      if ( defined $1 ) { say 5, "> disconnecting group $1"; } else { say 5, "> disconnecting all areas"; }
      while ( my ($tag, $area) = each %areas ) {
        if ( defined $1 && ($area->{group} ne $1) ) { next; }
        my $mode = area_node($area, $msg{'afrom'});
        if ( defined $mode ) {
          modify_link($tag, '-'.addr2s(@{$msg{'afrom'}}));
          $reply .= "$tag: disconnected\n";
        }
      }
    } 
    # -<area|mask>
    elsif ( /^\s*[-](\S+)/ ) {
      my $tag = uc $1;
      say 5, "> disconnecting area(s) $tag";
      my $s = do_unlink($tag, \%msg);
      if (defined $s) { $reply .= $s; }
      else { $reply .= "$tag: no matches\n"; }
    }
    # [+]<area|mask>
    elsif ( /^\s*[+]?(\S+)/ ) {
      my $tag = uc $1;
      say 5, "> connecting area(s) $tag";
      my $s = do_link($tag, \%msg);
      if (!defined($s) && ($tag !~ /[*?]/)) { 
        my $frq = $::FRQ_ENABLED ? forward_request($tag, $msg{'afrom'}) : undef;
        if (defined $frq) { $reply .= "$tag: $frq\n"; }
        else { $reply .= "$tag: unknown fileecho\n"; }
      }
      elsif (defined $s) { $reply .= $s; }
      else { $reply .= "$tag: no matches\n"; }
    }
    # unknown command
    else {
      say 5, "> unknown command at line $line: $_";
      $reply .= "$_: unknown command, skipping\n";
    }
  }
  # send reply msg
  if ($reply ne '') {
    $repl{'text'} = $reply."\n---\n";
    #Msg::fwrite(\%repl, $config{mailout}.'/');
    send_mail(\%repl);
  }
}
# --------------------------------------------------------------------
sub read_stdin
# process local request from stdin
{
  $| = 1;
  my @addr = s2addr $::vars{'addr'};
  if ( !defined $addr[0] ) { say 2, "Can't parse address '$::vars{addr}'"; return; }
  my $sfrom = addr4s(@addr);
  my %node = %{ $nodes{$sfrom} };
  if ( !%node ) { say 2, "Node $sfrom is not defined in config files"; return; }
  my @req = <STDIN>; $| = 1; print "\n";
  my %msg = ('afrom'=>\@addr);
  $msg{'from'} = ( defined $node{'sysop'} ) ? $node{'sysop'} : 'Sysop';
  say 2, "Processing local AreaManager request for $sfrom";
  mgr( join("\n", @req), \%msg );
}
# --------------------------------------------------------------------
# scans incoming mail folder for ffix requests
sub scan_mail
{
  my $msgno = 0;
  my $robots = join("\n", @{$config{'robot'}});
  while ( my %msg = Msg::nread($msgno, $config{mailin}.'/') ) {
    $msgno = $msg{'N'};
    # check destination address and robot name
    if ( ($robots =~ /^$msg{'to'}$/im) && my_aka(@{$msg{'ato'}}) ) {
      my $sfrom = addr4s(@{$msg{'afrom'}});
      say 2, "Processing AreaManager request from $sfrom";
      # check node record
      my %node = %{ $nodes{$sfrom} };
      if ( !%node ) {
        say 2, "> node $sfrom is not allowed to make AreaManager requests";
        # send error message
        send_mail( { from=>$ROBOT, afrom=>use_aka(@{$msg{afrom}}),
                     to=>$msg{from}, ato=>$msg{afrom},
                     subj=>'Failed request',
                     flags=>$Msg::PVT | $Msg::LOC,
                     text=>$Mgr::unlisted 
                   } );
        # del msg
        Msg::delete($msgno, $config{mailin}.'/');
        next;
      }
      # check password
      if ( (uc $msg{'subj'} ne uc $node{'robot_pwd'}) && !defined($::vars{'no_passwd'}) ) {
        say 2, "> wrong password from $sfrom: '$msg{subj}', expected '$node{robot_pwd}'";
        # send error message
        send_mail( { from=>$ROBOT, afrom=>$node{use_aka},
                     to=>$msg{from}, ato=>$msg{afrom},
                     subj=>'Failed request',
                     flags=>$Msg::PVT | $Msg::LOC | flag_msg($node{flag_mail}),
                     text=>$Mgr::wrong_pwd 
                   } );
        # del msg
        Msg::delete($msgno, $config{mailin}.'/');
        next;
      }
      # process
      mgr($msg{'text'}, \%msg);
      Msg::delete($msgno, $config{mailin}.'/');
    }
  }
}

# --------------------------------------------------------------------
# return integer value for char flag ('h','d','i','c')
sub flag_msg
{
  my %FV = (h=>$Msg::HLD, c=>$Msg::CRA, i=>$Msg::IMM, d=>$Msg::DIR);
  return $FV{lc $_[0]};
}

1;
