use POSIX qw(tmpnam);
use fidoaddr;
#use config;

# ====================================================================
# UTILITIES
# --------------------------------------------------------------------
# function log(int $lvl, string $s)
#    print string $s to log and/or console if it's verbosity <= $lvl
sub say
{
  my @l = localtime;
  my $dt = sprintf("%2d.%02d.%02d %2d:%02d ", $l[3], $l[4]+1, $l[5]%100, $l[2], $l[1]);
  if ($config{'loglevel'} >= $_[0] && defined LOG) { print LOG $dt.$_[1]."\n"; }
  if ($config{'conlevel'} >= $_[0] && !$vars{'quiet'}) { print $dt.$_[1]."\n"; }
}
# --------------------------------------------------------------------
# some functions to manage string data
sub first_word { my ($s) = $_[0] =~ /^\s*(\w+)/; return $s; }
sub last_word  { my ($s) = $_[0] =~ /(\w+)\s*$/; return $s; }
sub first_line { my ($s) = $_[0] =~ /([^\n]*)/s; return $s; }
sub rest_lines { my ($s) = $_[0] =~ /[^\n]*\n(.*)/s; return $s; }
# --------------------------------------------------------------------
# flag convertions
sub flag2s
{
  my %fs = (h=>'Hold', d=>'Direct', c=>'Crash', i=>'Immediate', o=>'Normal', f=>'Normal');
  return $fs{lc $_[0]};
}
# --------------------------------------------------------------------
# function run_crc32($s)
#    calc crc32 on $s
sub run_crc32
{
  if ( defined $config{'crc32'} ) {
    my $tmp_name = POSIX::tmpnam();
    open TMP, ">$tmp_name";
    print TMP $_[0];
    close TMP;
    my $crc32 = `$config{crc32} $tmp_name`;
    unlink $tmp_name;
    return hex $crc32;
  } 
  else { 
    require crc32;
    return crc32::crc32($_[0], length $_[0], -1); 
  }
}
# --------------------------------------------------------------------
# function file_crc32($name)
#    calc crc32 of file $name contents
sub file_crc32
{
  my $crc;
  # look in cache
  if ($config{'crc_cache'} && defined($crc_cache{$_[0]})) { return $crc_cache{$_[0]}; }
  # calc
  if ( defined $config{'crc32'} ) { my $crc = hex `$config{crc32} $_[0]`; } 
  else { 
    require crc32;
    $crc = 0; my $s;
    open F, $_[0]; binmode F;
    while (my $cnt = read F, $s, 16384) { $crc = crc32::crc32($s, $cnt, ~$crc); }
    close F;
  }
  if ($config{'crc_cache'}) { $crc_cache{$_[0]} = $crc; }
  return $crc; 
}
# ====================================================================
# ADDRESSED AND LINKS
# --------------------------------------------------------------------
# check if address is one of my aka's
sub my_aka
{
  if ( acmp($config{'address'}, \@_) ) { return 1; }
  my %akas = %{$config{'aka'}};
  while ( my ($k, $v) = each %akas ) {
    my @a = s2addr($k);
    if ( acmp(\@a, \@_) ) { return 1; }
  }
  return 0;
}
# --------------------------------------------------------------------
# return aka to use for node
sub use_aka
{
  my %akas = %{$config{'aka'}};
  while ( my ($k, $v) = each %akas ) {
    my @aka = s2addr($k);
    if ( acmp($v, \@_) ) { return \@aka; }
  }
  my @aka = @{$config{'address'}};
  return \@aka;
}
# --------------------------------------------------------------------
# convert array representation of links to string
sub make_links
{
  my $tag = $_[0];
  my $s = '';
  if ( @{$areas{$tag}{'links'}} > 0 ) {
    foreach $l (@{$areas{$tag}{'links'}}) {
      $s .= $l->{'mode'}.addr2s(@{$l->{'addr'}})." ";
    }
    chop $s;
  }
  return $s;
}
# --------------------------------------------------------------------
# return node mode in area record, undef if node is not a link
sub area_node
{
  my $area = $_[0];
  my $afrom = $_[1];
  say 8, "area_node($area, @$afrom)";
  if ( @{$area->{'links'}} > 0 ) {
    foreach $l (@{$area->{'links'}}) {
      if ( acmp($afrom, $l->{'addr'} ) == 1 ) { return $l->{'mode'}; }
    }
  }
  return undef;
}
# --------------------------------------------------------------------
# return 1 if area is avail to $node in $mode, 0 otherwise
sub area_avail
{
  my $tag = $_[0];
  my $afrom = $_[1];
  my $mode = (@_ > 2) ? $_[2] : ' ';
  my %area = %{$areas{$tag}};
  return 0 if ( !%area );
  my $group = $area{'group'};
  my %node = %{$nodes{ addr4s(@$afrom) }};
  return 0 if ( !%node );
  if ( $node{'groups'} =~ /$group/ ) { return 1; } else { return 0; }
}
# --------------------------------------------------------------------
# find domain for address @addr; undef for not found
sub domain
{
  if (@_ == 5) { return $_[4]; }
  my ($z, $n, $f, $p) = @_;
  return undef if ($z < 0);
  # if zone is the same as in main address, use outbound
  if ($z == $config{'address'}[0] && @{$config{'address'}} == 5) { return $config{'address'}[4]; }
  # scan aka's
  foreach my $k (keys %{$config{'aka'}}) {
    my @a = s2addr($k);
    if ( $z == $a[0] && @a == 5 ) { return $a[4]; }
  }
  return undef;
}
# --------------------------------------------------------------------
# function addr2bso(array @addr)
#    return path to outbound dir for @addr, domain-aware
sub addr2bso
{
  $PAT_BSO = '';
  if (@_ < 3) { return undef; }
  my ($z, $n, $f, $p) = @_; my $dom = domain(@_);
  # get outbound directory from domain
  my $outb;
  foreach my $arr (@{$config{'outb_dir'}}) {
    my($path, @addr) = @$arr;
    if ( acmp(\@_, \@addr) ) { $outb = $path; last; }
  }
  if (!defined($outb) && defined($dom) && lc($dom) ne lc($config{'address'}[4])) { ($outb) = $dom =~ /^([^.\/\\]*)/; }
  if (!defined($outb) || $outb eq basename($config{'outbound'})) { $outb = $PAT_BSO = $config{'outbound'}; }
  else { 
      $_ = $config{'outbound'}.'/..';
      $PAT_BSO = $_.'/'.pat_bso($outb);
      $outb = "$_/$outb";
    }
  if ($z != $config{'address'}[0]) { 
    my $ext = sprintf(".%03x", $z); $outb .= $ext; $PAT_BSO .= pat_bso($ext); 
  }
  my $d = '';
  if ($p == 0) { $outb .= $d = sprintf("/%04x%04x.", $n, $f); }
    else { $outb .= $d = sprintf("/%04x%04x.pnt/%08x.", $n, $f, $p); }
  $PAT_BSO .= pat_bso($d);
  return $outb;
}
# --------------------------------------------------------------------
# return case-independent pattern for glob(): 'a'->'[Aa], 'A'->'[Aa]'
sub pat_bso
{
  my $s = '';
  foreach my $ch (split //, $_[0]) { $s .= ($ch =~ /^[A-Za-z]$/) ? '['.uc($ch).lc($ch).']' : $ch; }
  return $s;
}
# --------------------------------------------------------------------
# function int32(int $number): return 32-radix representation of $number
sub int32
{
  my @dig = split //, '0123456789ABCDEFGHIJKLMNOPQRSTUV';
  my $i = $_[0]; my $s = '';
  while ($i > 0) { $s = $dig[$i%32].$s; $i = int($i/32); }
  return $s;
}
# --------------------------------------------------------------------
# function addr2box(\@addr, \%node)
#     returns filebox name for @addr, node record %node
#     undef means using bink-style outbound
sub addr2box
{
  my @addr = @{$_[0]}; my $node = $_[1];
  my $box = $node->{'box'};
  if (!defined $box) { $box = $config{'default_box'}; }
  if (defined($config{'t-mailbox'}) && ($box == 1)) {
    return $config{'t-mailbox'}.'/'
           .sprintf("%02s%03s%03s.%02s", int32($addr[0]), int32($addr[1]), int32($addr[2]), int32($addr[3]))
           .($node->{'flag_att'} eq 'h' ? 'H' : '');
  }
  elsif (defined($config{'t-maillongbox'}) && ($box == 2)) {
    return $config{'t-maillongbox'}.'/'
           .join('.', @addr[0..3]).($node->{'flag_att'} eq 'h' ? '.h' : '');
  }
  elsif (defined($config{'thebrake!box'}) && ($box == 3)) {
    my %flavour = ('h'=>'hold', 'c'=>'crash', 'i'=>'immediate', 'd'=>'direct', 'f'=>'normal');
    my $domain = domain(@addr);
    if (defined $domain) {
      $domain =~ s/[.].*$//;
      return $config{'thebrake!box'}.'/'
             .$domain.'.'.join('.', @addr[0..3]).'.'.$flavour{$node->{'flag_att'}};
    }
  }
  return undef;
}
# ====================================================================
# MAIL WRAPPER FUNCTIONS
# --------------------------------------------------------------------
# function send_mail(hash \%msg) - send netmail message
sub send_mail
{
  # .pkt
  if ( defined $config{'mailout_pkt'} ) {
    Pkt::write($_[0], $config{'mailout'}, $config{'mailout_pkt'});
  }
  # .msg
  else { Msg::fwrite($_[0], $config{'mailout'}.'/'); }
}
# --------------------------------------------------------------------
# function send_echo(hash \%msg) - send echomail message
sub send_echo
{
  my $echo = $_[0]->{'area'};
  my %echodefs = %{$config{'echoout'}};
  if ( !defined $echo ) { return 0; }
  while ( my ($echomask, $h) = each %echodefs ) {
    if ( $echo =~ /^$echomask$/i ) {
      if ( $h->{'type'} eq 'pkt' ) { Pkt::write($_[0], $h->{'path'}, $h->{'pkt'}); }
      elsif ( $h->{'type'} eq 'msg' ) { Msg::fwrite($_[0], $h->{'path'}.'/'); }
    }
  }
  return 1;
}
# ====================================================================
# AREA MANAGEMENT
# --------------------------------------------------------------------
# function find_descr($tag)
#    Look for a fileecho description
sub find_descr
{
  my $tag = $_[0];
  # for each fileecho list
  foreach my $v (@{$config{'fecholist'}}) {
    # seek files
    if (lc $v->[0] ne lc '@AreaLists@') {
      my $vv = $v->[0];
      my $ff = $v->[1]; $ff =~ s/[(][)]/($tag)/;
      if (!file_name_is_absolute($vv)) { $vv = rel2abs($vv, $BASE_PATH); }
      my @lists = glob $vv;
      @lists = sort { -M $a <=> -M $b } @lists;
      foreach my $list (@lists) {
        say 5, "> looking for descr in $list";
        if (!open LST, $list) { next; }
        my $s = join '', <LST>;
        close LST;
        if ($s =~ /$ff/mi) { say 5, "> descr found: '$2'"; return $2; }
      }
    }
    # seek via forwards
    else {
      my @feeds = forward_nodes('*');
      foreach my $arl (@feeds) {
        say 5, "> looking for descr in $arl->{file}";
        my @s = forward_areas($arl, $tag);
        if (defined $s[1]) { say 5, "> descr found: '$s[1]'"; return $s[1]; }
      }
    }
  }
  return undef;
}
# --------------------------------------------------------------------
# function create_area(string $tag, string $group, string $node[, \%keys])
#    Create area in memory, config file and its directory
sub create_area
{
  my ($tag, $group, $node) = @_;
  my $keys = $_[3] if @_ > 3;
  say 2, "Auto-creating area '$tag' from $node in group $group";
  # try to find corresponding default area
  my %area = ( defined $areas{'?'.$group} ) ? %{$areas{'?'.$group}} : %{$areas{'?'}};
  if ( !%area ) { say 2, "> can't auto-add area '$tag' from $node: no default area templates"; return 0; }
  # fetch description from TIC or look for it
  if (defined $keys->{'areadesc'}) { $area{'descr'} = $keys{'areadesc'}; }
  else { $area{'descr'} = find_descr($tag); }
  # set group
  if ( !defined $areas{'?'.$group} ) { $area{'group'} = $group; }
  # set creator/feed
  my @addr = s2addr($node);
  push @{$area{'links'}}, {'addr'=>\@addr, 'mode'=>'*'};
  # link all AutoConnect nodes for this area
  while ( my ($na, $nr) = each %nodes ) {
    if ( !$nodes{$na}{'autoconnect'}{$tag} ) { next; }
    say 3, "> connecting $na by AutoConnect";
    my @addr = s2addr($na);
    push @{$area{'links'}}, {'addr'=>\@addr, 'mode'=>''};
    # delete AutoConnect from config
    delete $nodes{$na}{'autoconnect'}{$tag};
    $updates{$na}{'del'}{'autoconnect'} = 1;
    my $s;
    foreach my $t (keys %{$nodes{$na}{'autoconnect'}}) { $s .= "AutoConnect $t\n"; }
    delete $updates{$na}{'add'}{'autoconnect'};
    $updates{$na}{'add'}{'autoconnect'} = $s if defined $s;
    # forward-requested area are pass-through
    delete $area{'path'}; delete $area{'mode'};
  }
  # set path and mkdir
  if (defined $area{'path'}) {
    my $path = $area{'path'};
    #my $uc_tag = uc $tag; $path =~ s/[$]A/$uc_tag/;
    #my $lc_tag = lc $tag; $path =~ s/[$]a/$lc_tag/;
    $path = parse_symb($path, '$', {'A'=>uc $tag, 'a'=>lc $tag});
    if ( -d $path ) { say 3, "> warn: directory $path already exists, using it"; }
    elsif (!mkdir $path, $area{'mode'} | 0111) {
      say 2, "> can't mkdir $path, auto-add failed for '$tag' from $node"; return 0;
    }
    $area{'path'} = $path;
  }
  # add to %areas, %newareas and @update_adds
  push @update_adds, $tag;
  $newareas{$tag} = [$area{'descr'}, \@addr];
  %{$areas{$tag}} = %area;
  # do OnCreate action
  if ( defined $area{'oncreate'} ) {
    my $n = (defined $keys{'fullname'}) ? $keys{'fullname'} : $keys{'file'};
    my $D = join("\n", @{$keys{'desc'}});
    my $d = (@{$keys{'ldesc'}}) ? join("\n", @{$keys{'ldesc'}}) : $D;
    my $cmd = ::parse_symb($area{'oncreate'}, '$', {
                             'A'=>$tag, 'a'=>lc $tag,
                             'F'=>$node, 'f'=>$node,
                             'G'=>$area{'group'}, 'g'=>$area{'group'},
                             'P'=>$area{'path'}, 'p'=>$area{'path'}
                           });
    ::say 5, "> executing command '$cmd'";
    $res = system $cmd;
    ::say 5, "> execution finished, exit code=".($res>>8);
  }
  return 1;
}
# --------------------------------------------------------------------
# function unfeed_area($tag, $addr)
#    called when feed unsubscribes
sub unfeed_area
{
  my $tag = $_[0];
  my $public = 0; my $remove = 0;
  ::say 2, "Feed unsubscribes from area $tag";
  # send unlinks requests
  if ( $config{'feedunlinks'} =~ /public/ ) {
    foreach $l (@{$areas{$tag}{'links'}}) {
      my $to = addr4s(@{$l->{'addr'}});
      if ( !defined $nodes{$to}{'robot'} ) { next; }
      ::say 3, "> sending unsubscribe request to $to";
      send_mail( { from=>$SYSOP, afrom=>$nodes{$to}{'use_aka'},
                   to=>$nodes{$to}{'robot'}, ato=>$l->{'addr'},
                   subj=>$nodes{$to}{'robot_pwd'},
                   flags=>$Msg::PVT | $Msg::LOC,
                   text=>"-$tag\n" 
                 } );
    }    
    $public = 1;
  }
  # removing area from config
  if ( $config{'feedunlinks'} =~ /remove/ ) {
    ::say 3, "> removing area $tag from configs";
    delete $updates{$tag};
    $updates{$tag}{'del'}{'*'} = 1;
    delete $areas{$tag};
    $remove = 1;
  }
  # notify sysop
  if ( $config{'feedunlinks'} !~ /silent/ ) {
    my $s;
    if ( $remove ) { $s .= ", area is removed from configs"; }
      else { $s .= ", area is keeped in configs"; }
    if ( $public ) { $s .= ", unlink requests sent to all links"; }
    send_mail( { from=>$ROBOT, afrom=>$config{'address'},
                 to=>$SYSOP, ato=>$config{'address'},
                 subj=>'Feed unsubscribes from area $tag',
                 flags=>$Msg::PVT | $Msg::LOC,
                 text=>"Feed ".addr2s(@{$_[1]})." unsubscribes from area $tag$s\n"
               }
             );
  }
}
# --------------------------------------------------------------------
# function empty_area($tag[, $last])
#    called when last link unsubscribes from pass-through area
#    $last is 1 if the last unsubscribes, 0 if the last remains
sub empty_area
{
  my $tag = $_[0]; my $last = $_[1];
  my $remove = 0; my $public; my $to; my $reason;
  if ($last) {
    $reason = "Last link unsubscribes from";
    ::say 2, $reason." pass-through area $tag";
  }
  elsif ($config{'lastunlinks'} !~ /strict/) {
    $reason = "Only one link remains in";
    ::say 2, $reason." pass-through area $tag";
    # send unlinks requests
    $public = 0;
    my @addr = @{$areas{$tag}{'links'}[0]{'addr'}};
    $to = addr4s(@addr);
    if ( $config{'lastunlinks'} =~ /public/ ) {
      if ( defined $nodes{$to}{'robot'} ) {
        ::say 3, "> sending unsubscribe request to $to";
        send_mail( { from=>$SYSOP, afrom=>$nodes{$to}{'use_aka'},
                     to=>$nodes{$to}{'robot'}, ato=>\@addr,
                     subj=>$nodes{$to}{'robot_pwd'},
                     flags=>$Msg::PVT | $Msg::LOC,
                     text=>"-$tag\n" 
                   } );
      }
      $public = 1;
    }
  }
  else { return; }
  # removing area from config
  if ( $config{'lastunlinks'} =~ /remove/ ) {
    ::say 3, "> removing area $tag from configs";
    delete $updates{$tag};
    $updates{$tag}{'del'}{'*'} = 1;
    delete $areas{$tag};
    $remove = 1;
  }
  # notify sysop
  if ( $config{'lastunlinks'} !~ /silent/ ) {
    my $s;
    if ( $remove ) { $s .= ", area is removed from configs"; }
      else { $s .= ", area is keeped in configs"; }
    if (defined $public) {
      if ( $public ) { $s .= ", unlink request sent to $to"; }
      else { $s .= ", unlink request is NOT sent to $to"; }
    }
    send_mail( { from=>$ROBOT, afrom=>$config{'address'},
                 to=>$SYSOP, ato=>$config{'address'},
                 subj=>$reason.' area $tag',
                 flags=>$Msg::PVT | $Msg::LOC,
                 text=>$reason." pass-through area $tag$s\n"
               }
             );
  }
}
# ====================================================================
# FORWARD REQUESTS
# --------------------------------------------------------------------
# function forward_nodes($addr)
#    get a list of arealist records from nodes, accepting forwreqs
#    or get a list of all available arealist records if $addr eq '*'
sub forward_nodes
{
  my $addrs = $_[0];
  # scan all nodes with arealists
  my @feeds = ();
  while ( my ($a, $n) = each %nodes ) {
    if (!$n->{'arealist'}) { next; }
    # use special '*' value for $addrs to get 
    if ($addrs ne '*') {
      # prevent forwreqs with 'off' flag
      if ($n->{'arealist'}{'off'}) { next; }
      # check if target node can add areas and has defined robot name
      if (!defined $n->{'auto_add'} || !$n->{'robot'}) { next; }
      # check if requestor node can connect to auto-create group of feed
      my $groups = $nodes{$addrs}{'groups'}; 
      my $aagrp  = $n->{'auto_add'};
      if (!defined $aagrp || ($groups !~ /$aagrp/)) { next; }
    }
    # if permissions allow, add to candidates array
    my %h = %{$n->{'arealist'}}; $h{'addr'} = $a;
    push @feeds, \%h;
  }
  @feeds = sort { $a->{'prio'} <=> $b->{'prio'} } @feeds;
  return @feeds;
}
# --------------------------------------------------------------------
# function forward_areas($arl[, $tag])
#    return array of available areas at node if $tag undefined
#    return 1 if area $tag is present in file
sub forward_areas
{
  my $arl = $_[0]; my $tag = $_[1]; my @list = ();
  # check if arealist file exists
  my $file = $arl->{'file'};
  if (!-r $file) {
    $file = rel2abs($file, $::BASE_PATH);
    if (!-r $file) { return @list; }
  }
  # read areas list and find area
  if (open LST, $file) { 
    my @arr = <LST>; close LST; 
    # make pattern
    my $pat = $arl->{'format'}; 
    $pat = '\s*(\S+)(?:\s+(.*?)\s*)?$' if !defined $pat;
    if (defined $tag) { $pat =~ s/\([^)]*/\($tag/; }
    # do search
    foreach my $s (@arr) {
      if ( defined $tag ) {
#        if ($s =~ /^\s*$tag\s+/mi) { return 1; }
        if ($s =~ /$pat/mi) { return (uc $1, $2); }
      }
      else {
#        if ($s =~ /^\s*(\S+)\s+(.*?)\s*$/mi) { push @list, [uc $1, $2]; }
        if ($s =~ /$pat/mi) { push @list, [uc $1, $2]; }
      }
    }
  } 
  return (defined $tag) ? 0 : @list;
}
# --------------------------------------------------------------------
# function forward_request($tag, $addr)
#    called any time when unknown area requested, return undef if
#    no matching areas found, else return answer to mail
sub forward_request
{
  my $tag = $_[0]; my @addr = @{$_[1]}; my $addrs = addr4s(@addr);
  # check for 'forwreqs' in options
  if ($nodes{$addrs}{'options'} !~ /forwreqs/i) { return undef; }
  # check if already connected
  if ($::forward_reqs{$tag}) {
    if (!defined $nodes{$addrs}{'autoconnect'}{$tag}) {
      $nodes{$addrs}{'autoconnect'}{$tag} = 1;
      $updates{$addrs}{'add'}{'autoconnect'} .= "AutoConnect $tag\n";
    }
    return "request already forwarded, waiting for files"; 
  }
  # get list of forward-to nodes
  my @feeds = forward_nodes($addrs);
  foreach my $arl (@feeds) {
    # check for uncond
    if ( ! $arl->{'uncond'} ) {
      if ( ! forward_areas($arl, $tag) ) { next; }
    }
    # if found area or enabled unconditional reqs - send a request
    my @addr = s2addr($arl->{addr});
    my $robot = $nodes{$arl->{addr}}{'robot'};
    my $robot_pwd = $nodes{$arl->{addr}}{'robot_pwd'};
    send_mail( { from=>$::SYSOP, afrom=>use_aka(@addr),
                 to=>$robot, ato=>\@addr,
                 subj=>$robot_pwd, flags=>$Msg::PVT|$Msg::LOC,
                 text=>"+$tag\n"
               }
             );
    $updates{$addrs}{'add'}{'autoconnect'} .= "AutoConnect $tag\n";
    return "request forwarded to $arl->{addr}"; 
  }
  return undef;
}
# --------------------------------------------------------------------
# function forward_avail($addr)
#    return available areas list for node $addr
sub forward_avail
{
  my @addr = @{$_[0]}; my $addrs = addr4s(@addr);
  my $aa = use_aka(@addr); $Tpl::aka = addr2s(@$aa);
  # check for 'forwreqs' in options
  if ($nodes{$addrs}{'options'} !~ /forwreqs/i) { 
    $Tpl::allowed = 0; return do 'tpl/avail.pl'; 
  }
  # get sorted list of forward-to nodes
  my @feeds = forward_nodes($addrs);
  # get sorted list of areas
  my @lst = ();
  foreach my $arl (@feeds) {
    if ( $arl->{'hidden'} ) { next; }
    my @tmp = forward_areas($arl);
    @tmp = sort { $a->[0] cmp $b->[0] } @tmp;
    push @lst, { 'addr'=>$arl->{addr}, 'list'=>\@tmp };
  }
  # return
  @Tpl::avail = @lst; $Tpl::allowed = 1;
  do 'tpl/avail.pl';
  return @Tpl::res;
}
# ====================================================================
# STRING PARSING
# --------------------------------------------------------------------
# function parse_symb(string $s, char $ch, hash \%repl)
#    parse $s, $ch is special '\'-style char, replaces are in %repl
sub parse_symb
{
  my $S = 0; my @out = (); my %R = %{$_[2]};
  my @in = split //, $_[0];
  foreach my $w (@in) { 
    if ($w eq $_[1] && !$S) { $S = 1; next; }
    if ($S) {
      if ( defined $R{$w} ) { $w = $R{$w}; }
      elsif ( $w ne $_[1] ) { $w = $_[1].$w; }
    }
    push @out, $w; $S = 0;
  }
  return join('', @out);
}
# --------------------------------------------------------------------
# function parse_quote(string $s[, char $ch])
#    parse $s, as string in $ch quotes (or to end of line if $ch omitted)
sub parse_quote
{
  my $S = 0; my @out = (); my $D = undef; my $s = $_[0];
  if (@_ > 1 && $s =~ /^([$_[1]])/) { $D = $1; }
  my @in = split //, $s; shift @in if defined $D;
  foreach my $w (@in) { 
    if ($w eq '\\' && !$S) { $S = 1; next; }
    if ($w eq $D && !$S) { last; }
    if ($S) { $w = ($w eq '"') ? '"' : eval("\"\\$w\""); }
    push @out, $w; $S = 0;
  }
  return join('', @out);
}
# the same as above, but $ch is required; return unchanged line if not $ch
sub parse_quote2
{
  my $S = 0; my @out = (); my $D = undef; my $s = $_[0];
  if (@_ == 2 && $s =~ /^([$_[1]])/) { $D = $1; } else { return $_[0]; }
  my @in = split //, $s; shift @in;
  foreach my $w (@in) { 
    if ($w eq '\\' && !$S) { $S = 1; next; }
    if ($w eq $D && !$S) { last; }
    if ($S) { $w = ($w eq '"') ? '"' : eval("\"\\$w\""); }
    push @out, $w; $S = 0;
  }
  return join('', @out);
}
# --------------------------------------------------------------------
# function parse_cmdline(string $cmd_line, hash \%opt, hash \%var)
#    parse command line to find options in %opt, get them to %var
sub parse_cmdline
{
  $_ = ' '.$_[0];
  my %opt = %{$_[1]};
  my $var = $_[2];
  study $_;
  while ( my ($key, $arr) = each %opt ) {
    if ( !@$arr ) { $arr = [$key]; }
    foreach $opt (@$arr) {
      $opt =~ /(.*?)((?<!\\)[_=])?((?<!\\)[?])?$/;
      my ($s, $mod, $ns) = ($1, $2, $3);
      if (!$mod) { 
        if ( /\s+$s(\s+|\s*$)/ ) { $var->{$key} = 1; s/$s//;}
      } else {
        if ($mod eq '_') {
          if ( /\s+$s(?:\s+(\S+))$ns/ ) { $var->{$key} = $1.''; $_ = $`.$';
            #substr($_, $-[0], $+[0] - $-[0]) = ''; 
          }
        }
        elsif ($mod eq '=') {
          if ( /\s+$s(?:=(\S+))$ns/ ) { $var->{$key} = $1.''; $_ = $`.$';
            #substr($_, $-[0], $+[0] - $-[0]) = ''; 
          }
        }
      }
    }
  }
  s/^\s*(.*?)\s*$/$1/;
  return $_;
}
# ====================================================================
# TEMPLATE PARSING
# --------------------------------------------------------------------
# function tpl2s(string $tpl)
#    Make string from a template
#    a) line, not followed by /^`.*?'$/ or /^'.*?`$/ passed as-is
#    b) line, followed by /^'.*?`$/ passed as sprintf(<line>, eval $1)
#    c) line, followed by /^`.*?'$/ passed as formline(<line>, eval $1)
#       for each text line, contained in $1
sub tpl2s
{
  my $last = '';
  my $res = '';
  my @arr = split /\n/, $_[0];
  foreach (@arr) {
    if ( /^'(.*?)`$/ ) {
      my $vals = $1;
      $last = sprintf($last, eval($vals));
    }
    elsif ( /^`(.*?)'$/ ) {
      chomp $last;
      my $res = '';
      my @vals = eval($1);
      while (1) {
        my @curs = ();
        for (my $i = 0; $i < @vals; $i++) {
          my $s = $vals[$i];
          $s =~ /([^\n]*)\n?(.*)/s;
          $curs[$i] = $1; $vals[$i] = $2;
        }
        if ( join('', @curs) eq '' ) { last; }
        $^A = '';
        formline $last, @curs;
        $res .= $^A."\n"; 
        $^A = '';
      }
      $last = $res;
    }
    elsif ( $_ eq '.' ) { last; }
    else { $res .= $last; $last = $_."\n"; }
  }
  $res .= $last;
  return $res;
}
# ====================================================================
# TRAFFIC MANAGEMENT
# --------------------------------------------------------------------
# function update_traffic($tag[, $force])
#    update traffic for the current and last month for area $tag
#    if $force, create taffic record if it doesn't exist; else return undef
sub update_traffic {
  my $trf = $::traffic{uc $_[0]};
  if (!%$trf) {
    return undef unless $_[1];
    #$trf = {cmf=>-1, cmk=>-1, cwf=>-1, cwk=>-1};
    $trf = {};
    $::traffic{uc $_[0]} = $trf;
  }
  my ($uy, $um, $ud, $uwd) = (localtime $trf->{'upd'})[5,4,3,6];
  my ($cy, $cm, $cd, $cwd) = (localtime)[5,4,3,6];
  my $days_diff = (time - $trf->{'upd'}) / (60*60*24);
  my $mons_diff = ($cm - $um) + 12*($cy - $uy);
  my $new_week = ($cwd < $uwd) && ($cd > $ud || $cm > $um || $cy > $uy);
  my $new_month = ($cm > $um) || ($cy > $uy);
  if ($new_month) {
    $trf->{'lmf'} = $mons_diff > 1 ? 0 : $trf->{'cmf'}; $trf->{'cmf'} = 0;
    $trf->{'lmk'} = $mons_diff > 1 ? 0 : $trf->{'cmk'}; $trf->{'cmk'} = 0;
  }
  if ($new_week) {
    $trf->{'lwf'} = $days_diff > 7 ? 0 : $trf->{'cwf'}; $trf->{'cwf'} = 0;
    $trf->{'lwk'} = $days_diff > 7 ? 0 : $trf->{'cwk'}; $trf->{'cwk'} = 0;
  }
  $trf->{'upd'} = time;
  return $trf;
}

1;
