use POSIX qw(mktime strftime tmpnam);
use Time::Local;
use fidoaddr;
# --------------------------------------------------------------------
{
  package Msg;
  # message flags constants
  *PVT = \0x0001;
  *CRA = \0x0002;
  *RCV = \0x0004;
  *SNT = \0x0008;
  *ATT = \0x0010;
  *TRS = \0x0020;
  *ORP = \0x0040;
  *K_S = \0x0080;
  *LOC = \0x0100;
  *HLD = \0x0200;
  *RSV = \0x0400;
  *FRQ = \0x0800;
  *RRQ = \0x1000;
  *RRC = \0x2000;
  *ARQ = \0x4000;
  *URQ = \0x8000;
  *A_S = \0x00010000;
  *DIR = \0x00020000;
  *ZON = \0x00040000;
  *HUB = \0x00080000;
  *IMM = \0x00100000;
  *XMA = \0x00200000;
  *KFS = \0x00400000;
  *TFS = \0x00800000;
  *LOK = \0x01000000;
  *CFM = \0x02000000;
  # fax flags
  *HIR = \0x04000000;
  *COV = \0x08000000;
  *SIG = \0x10000000;
  *LET = \0x20000000;
  # flags in ^aFLAGS kludge
  %kl_flags = ('PVT'=>$PVT, 'CRA'=>$CRA, 'RCV'=>$RCV, 'SNT'=>$SNT,
               'FIL'=>$ATT, 'K/S'=>$K_S, 'HLD'=>$HLD, 'FRQ'=>$FRQ,
               'RRQ'=>$RRQ, 
               'A/S'=>$A_S, 'DIR'=>$DIR, 'ZON'=>$ZON, 'HUB'=>$HUB, 
               'IMM'=>$IMM, 'XMA'=>$XMA, 'KFS'=>$KFS, 'TFS'=>$TFS, 
               'LOK'=>$LOK, 'CFM'=>$CFM,
               'HIR'=>$HIR, 'COV'=>$COV, 'SIG'=>$SIG, 'LET'=>$LET);
  $path = '';
  # --------------------------------------------------------------------
  sub parse_date
  {
    my @mons = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun', 'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec');
    ($d,$m,$y,$H,$M,$S) = $_[0] =~ m/(\d{2}) (\w{3}) (\d{2})  (\d{2}):(\d{2}):(\d{2})/;
    for ($i = 0; $i < @mons; $i++) { if ($m eq $mons[$i]) { last; } }
  #  my ($d, $m, $y, $H, $M, $S) = ($1, $i+1, ($3>80)?$3+1900:$3+2000, $4, $5, $6);
  #  if ( $d < 1 || $d > 31 || $m < 1 || $m > 12 || $y < 1981 || $y > 2080 ||
  #       $H < 0 || $H > 23 || $M < 0 || $M > 59 || $S < 0 || $S > 59 ) return undef;
  #  print "mktime($S, $M, $H, $d, $i, ".(($y<80)?$y+100:$y).")=".POSIX::mktime($S, $M, $H, $d, $i, ($y<80)?$y+100:$y)."\n";
  #  return ::mktime($S, $M, $H, $d, $i, ($y<80)?$y+100:$y);
  #  print "timelocal($S, $M, $H, $d, ".($i).", ".(($y<80)?$y+2000:$y+1900).")=".timelocal($S, $M, $H, $d, $i, (($y<80)?$y+2000:$y+1900))."\n";
    return ::timelocal($S, $M, $H, $d, $i, (($y<80)?$y+2000:$y+1900));
  }
  # --------------------------------------------------------------------
  # nread(int $prev_msgno[, string $path])
  sub nread
  {
    my $N = $_[0]+0;
    $path = $_[1] if (@_ > 1);
    my $n = 0x10000;
    # get next number
    opendir DIR, $path;
    while ( my $name = readdir DIR ) {
      if ( $name =~ /(\d+)\.[Mm][Ss][Gg]/ ) { $n = $1 if ($1 > $N) && ($1 < $n) };
    }
    close DIR;
    return ($n < 0x10000) ? Msg::fread($n) : ();
  }
  # --------------------------------------------------------------------
  # fread(int $msgno[, string $path])
  sub fread
  {
    $path = $_[1] if (@_ > 1);
    if ($_[0] !~ /\.[Mm][Ss][Gg]$/) { $_[0] .= ".msg"; }
    # read file
    open F, $path.$_[0]; binmode F;
    read F, $_, -s F;
    close F;
    # parse and return
    return Msg::read($_, int $_[0]);
  }
  # --------------------------------------------------------------------
  sub read
  {
    my $msgid, $reply;
    $_ = $_[0];
    # read message header
    my ($from, $to, $subj, $date,
     $times_read, $to_node, $from_node, $cost,
     $from_net, $to_net, $to_zone, $from_zone,
     $to_point, $from_point, $reply, $flags, $answer, $txt) = unpack "Z36 Z36 Z72 Z20 s11 S1 s1 Z*", $_;
    # skip \n's and convert end-of-line (\r in msg, \n usual)
    $txt =~ s/\n//g;
    $txt =~ tr/\r/\n/;
    study $txt;
    # make addresses from header - can't rely on zones and points :(
    my @from = (-1, $from_net, $from_node, 0);
    my @to   = (-1, $to_net, $to_node, 0);
    # process ^aINTL
    if ( $txt =~ m/^\x01INTL (\S+) (\S+)$/m ) { @from = ::s2addr($2); @to = ::s2addr($1); }
    # ^aFMPT
    if ( $txt =~ m/^\x01FMPT (\d+)$/m ) { $from[3] = $1; }
    # ^aTOPT
    if ( $txt =~ m/^\x01TOPT (\d+)$/m ) { $to[3] = $1; }
    # ^aINTL:
    if ( $txt =~ m/^\x01INTL: (\S+) (\S+)$/m ) { @from = ::fqfa2addr($2); @to = ::fqfa2addr($1); }
    # ^aMSGID
    if ( $txt =~ m/^\x01MSGID: (?:"(.*?)"(?!")|(\S+)) (\S+)/m ) { 
      my @msgida = ::s2addr($1.$2);
      if ( defined @msgida && @msgida[1..2] == @from[1..2] ) { 
        $from[0] = $msgida[0] if $from[0] == -1;
        $to[0] = $msgida[0] if $to[0] == -1;
      }
      $msgid = "$1$2 $3";
    }
    # ^aREPLY
    if ( $txt =~ m/^\x01REPLY: (?:"(.*?)"(?!")|(\S+)) (\S+)/m ) { $reply = "$1$2 $3"; }
    # ^aFLAGS
    if ( $txt =~ m/^\x01FLAGS (.*?)$/m ) {
      $_ = $1;
      while ( my($n, $v) = each %Msg::kl_flags ) { if (/$n/) { $flags |= $v; } }
    }
    # make clean text (w/o kludges)
    my $text = $txt; $text =~ s/\x01[^\n]*\n//sg;
    # make hash as the message representation
    my %msg = ();
    my @afrom = @from; my @ato = @to;
    $msg{'N'} = $_[1]; $msg{'afrom'} = \@from; $msg{'ato'} = \@to; 
    $msg{'from'} = $from; $msg{'to'} = $to; $msg{'subj'} = $subj; 
    $msg{'time'} = parse_date $date; $msg{'flags'} = $flags;
    $msg{'txt'} = $txt; $msg{'text'} = $text;
    return %msg;
  }
  # --------------------------------------------------------------------
  # function write(hash \%msg[, "no_header"])
  #    for echomail must define $msg{'area'} - not to write INTL, FMPT, TOPT
  sub write
  {
    my %msg = %{$_[0]};
    # set undefined parameters
    $msg{'time'} = time if !defined $msg{'time'};
    # get message parameters
    my @afrom = @{$msg{'afrom'}}; my @ato = @{$msg{'ato'}};
    # create kludges
    my $kl;
    if ( !defined $msg{'area'} ) {
      $kl = "\x01INTL ".::addr2s( @ato[0..2] ).
            " ".::addr2s( @afrom[0..2] )."\n";
      if ( @ato[3] > 0 ) { $kl .= "\x01TOPT ".@ato[3]."\n"; }
      if ( @afrom[3] > 0 ) { $kl .= "\x01FMPT ".@afrom[3]."\n"; }
    }
    if ( !defined $msg{'msgid'} ) {
      my $s = pack 'L S S', $msg{'time'}, $msg{'N'}, $msg{'flags'};
      my $crc = ::run_crc32($s);
      $msg{'msgid'} = ::addr2s(@afrom)." ".sprintf("%08x", $crc);
    }
    $kl .= "\x01MSGID: $msg{msgid}\n";
    if ( defined $msg{'reply'} ) { $kl .= "\x01REPLY: $msg{reply}\n"; }
    # flags
    $fl = $msg{'flags'} & 0xFFFF0000; $sfl = '';
    while (my($k,$v) = each %kl_flags) {
      $sfl .= " $k" if ($fl & $v);
    }
    $kl .= "\x01FLAGS$sfl\n" if ($sfl ne '');
    # header
    $date = POSIX::strftime('%d %b %y  %H:%M:%S', localtime $msg{'time'});
    my $hdr = pack "Z36 Z36 Z72 Z20 s11 S1 s1", ($msg{'from'}, $msg{'to'}, $msg{'subj'}, $date,
      $msg{'times_read'}, $ato[2], $afrom[2], $msg{'cost'}, $afrom[1], $ato[1], 
      $ato[0], $afrom[0], $ato[3], $afrom[3], $msg{'reply_no'}, 
      $msg{'flags'} & 0xFFFF, $msg{'answer'}
    ) unless ($_[1] eq 'no_header');
    # i think, all done
    $_ = $kl.$msg{'text'}."\x00";
    tr/\n/\r/;
    return $hdr.$_;
  }
  # --------------------------------------------------------------------
  # fwrite(hash %msg[, string $path])
  sub fwrite
  {
    my %msg = %{$_[0]};
    # determine max msg-number
    $path = $_[1] if (@_ > 1);
    if (!defined $msg{'N'}) {
      my $max = 0;
      opendir DIR, $path;
      while ( my $name = readdir DIR ) {
        if ( $name =~ /(\d+)\.[Mm][Ss][Gg]/ ) { $max = $1 if $1 > $max };
      }
      close DIR;
      $msg{'N'} = $max + 1;
    }
    # create msg
    $_ = Msg::write(\%msg);
    # write to file
    open F, ">$path$msg{N}.msg"; binmode F;
    print F $_;
    close F;
  }
  # --------------------------------------------------------------------
  # delete($msgno[, $path])
  sub delete {
    $path = $_[1] if (@_ > 1);
    my $msg = $_[0]; if ($msg !~ /[.][Mm][Ss][Gg]$/) { $msg .= ".msg"; }
    unlink "$path/$msg";
  }
}

1;
