use POSIX qw(tmpnam strftime);
use fidoaddr;
use msg;
# --------------------------------------------------------------------
{
  package Pkt;
    $pkt_name;				# name of .pkt in process
    $pkt_type;				# .pkt type (20=fts-0001, 21=fsc-0039, 22=fsc-0048)
  # ------------------------------------------------------------------
  # function create(hash \%msg[, string $path]) - creates .pkt header
  sub create
  {
    # ensure closing
    if ( defined $pkt_name ) { Pkt::close(); }
    # some input
    my %msg = %{$_[0]};
    my @from = @{$msg{'afrom'}}; my @to = @{$msg{'ato'}};
    my @t = localtime; $t[5] %= 100;
    # get pkt name
    $pkt_name = (@_ > 1) ? $_[1].'/' : './';
    $pkt_type = (@_ > 2) ? $_[2] : 21;
    my $i = 0;
    while ( -f $pkt_name.sprintf("ffix%04x.pkt", $i) ) { $i++; }
    $pkt_name .= sprintf("ffix%04x.pkt", $i);
    # type-2 (fts-0001) header
    my $hdr;
    if ($pkt_type == 20) {
       $hdr = pack 'S12 C2 Z8 S2 Z18', $from[2], $to[2], 
                                       $t[5], $t[4], $t[3], $t[2], $t[1], $t[0],
                                       0, 2, $from[1], $to[1], 0xfe, 0, $msg{'password'}, $from[0], $to[0], '';
    } elsif ($pkt_type != 22) {
    # type-2 ext (fsc-0039) header
       $hdr = pack 'S12 C2 Z8 S2 L C2 S5 L', $from[2], $to[2], 
                                       $t[5], $t[4], $t[3], $t[2], $t[1], $t[0],
                                       0, 2, $from[1], $to[1], 0xfe, 0, $msg{'password'}, $from[0], $to[0],
                                       0, 0, 0, 0x0002, $from[0], $to[0], $from[3], $to[3], 0;
    } elsif ($pkt_type == 22) {
    # type-2+ (fsc-0048) header
       $hdr = pack 'S12 C2 Z8 S2 S2 C2 S5 L', $from[2], $to[2], 
                                       $t[5], $t[4], $t[3], $t[2], $t[1], $t[0],
                                       0, 2, ($from[3] ? -1 : $from[1]), $to[1], 0xfe, 0, $msg{'password'}, $from[0], $to[0],
                                       ($from[3] ? $from[1] : 0), 0x0200, 0, 0, 0x0002, $from[0], $to[0], $from[3], $to[3], 0;
    }
    open PKT, ">$pkt_name"; binmode PKT;
    print PKT $hdr, "\x00\x00";
  }
  # ------------------------------------------------------------------
  # function write(hash \%msg) - append an .msg to current .pkt
  sub write
  {
    my %msg = %{$_[0]};
    my @from = @{$msg{'afrom'}}; my @to = @{$msg{'ato'}};
    # create .pkt if not found
    if ( !defined $pkt_name || $pkt_name eq '' ) { create(@_); }
    # seek two bytes backward (null-terminator)
    seek PKT, -2, 2;
    # generate .msg
    my $s = Msg::write($_[0], 'no_header');
    # add packed message header
    my $hdr = pack 'S6 Z20', $from[2], $to[2], $from[1], $to[1], 
                             $msg{'flags'} & 0xFFFF, $msg{'cost'},
                             POSIX::strftime('%d %b %y  %H:%M:%S', localtime);
    $hdr .= substr($msg{'to'},   0, 35)."\x00";
    $hdr .= substr($msg{'from'}, 0, 35)."\x00";
    $hdr .= substr($msg{'subj'}, 0, 71)."\x00";
    if ( defined $msg{'area'} ) { $s = "AREA:$msg{area}\r$s"; }
    # write message
    print PKT "\x02\x00", $hdr.$s, "\x00\x00";
  }
  # ------------------------------------------------------------------
  # function close() - closes .pkt
  sub close
  {
    if ( defined $pkt_name ) { close PKT; undef $pkt_name; }
  }
  # ------------------------------------------------------------------
  sub END { Pkt::close(); }
}

#Pkt::write({'area'=>'TEST', 'afrom'=>[2,463,180,8], 'ato'=>[2,463,180,8], 'from'=>'val khokhlov', 'to'=>'val khokhlov', 'subj'=>'test', 'text'=>"aaa\nbbb\nccc\n---\n"});

1;

#sub run_crc32 { return '000000'; }
