use File::Copy qw(copy move);
use fidoaddr;
use config;
use misc;

{
  package Tic;
  # ------------------------------------------------------------------
  # error codes
  *OK   =  \0;
  *BAD  = \-1;
  *DUPE = \-2;
  # processing flags for Tic::process
  *NO_COPY = \0x01;
  *NO_DEL  = \0x02;
  *NO_REPL = \0x04;
  *NO_FBBS = \0x08;
  *NO_DUPE = \0x10;
  # keywords mask: 1 - mandatory keyword, 2 - not pass-through, 4 - multiple
  %keywords = ('area'=>1, 'areadesc'=>0, 'file'=>1, 'fullname'=>2, 'crc'=>1, 
               'magic'=>2, 'replaces'=>0, 'desc'=>4, 'ldesc'=>4, 'size'=>0, 
               'date'=>0, 'release'=>0, 'author'=>0, 'source'=>0, 'app'=>4,
               'origin'=>0, 'from'=>3, 'to'=>2, 'created'=>3, 'via'=>4, 
               'path'=>5, 'seenby'=>5, 'pw'=>1, 'pgp'=>0, 'receiptrequest'=>2
              );
#  @keywords = ('area', 'areadesc', 'origin', 'from', 'to', 'replaces', 'magic',
#               'file', 'desc', 'ldesc', 'crc', 'size', 'date', 'release',
#               'author', 'source', 'app', 'created', 'via', 'path', 'seenby',
#               'pw', 'pgp', 'receiptrequest'
#              );
  # ------------------------------------------------------------------
  %area;                                # current .tic's area
  #%node;                                # current .tic's sender
  %keys;                                # incoming .tic's keys
  @ttic;                                # unknown .tic's keys - transit
  $file;                                # file for current .tic
  @in_dirs;				# inbound directories
  @in_files;				# inbound listing
  # ------------------------------------------------------------------
  sub read_keys 
  {
    $_ = $_[0];
    # conver <cr><lf> pair to current <eol>
    s!\r\n!$/!g;
    # parse tic line-by-line
    undef %keys; undef @ttic;
    my @lines = split /$\//;
    foreach (@lines) {
      my $v = '';
      ($k, $v) = /^(\S+)\s?(.*?)$/;
      $k =~ tr/[A-Z]/[a-z]/;
      if ( defined $keywords{$k} ) {
        $t = $keywords{$k};
        if ($t & 4) { push @{$keys{$k}}, $v; } else { $keys{$k} = $v; }
        # don't need since tic is re-created by me
        #if ( !($t & 2) && !($v eq '') ) { push @ttic, $_; }
      } 
      # push unknown by me keywords
      else { push @ttic, $_; }
    }
  }
  # ------------------------------------------------------------------
  sub desc_convert
  {
    my ($i, $k);
    my ($s) = @_;
    return () if length $s < 64;
    for ($i = 20; $i < 64; $i++) {
      $k = 1;
      while ($k*$i+$k-1 < length $s) {
        if ( substr($s, $k*$i+$k-1, 1) ne ' ') { $k = 0; last; } else { $k++; }
      }
      last if $k;
    }
    # array: @a = $s =~ /(.{1,$i})\s?/g;
    # string: $s =~ s/(.{1,$i})\s?/\1/;
    return ($i < 64) ? ($s =~ /(.{1,$i})\s?/g) : ();
  }
  # ------------------------------------------------------------------
  sub find_by_crc
  {
    my $i;
    if (!@in_files) { foreach $i (@in_dirs) { push @in_files, glob "$i/*"; } }
    foreach $i (@in_files) {
      if (defined $_[1]) { next unless (-s $i == $_[1]); }
      if (::file_crc32($i) == hex $_[0]) { return $i; }
    }
    return undef;
  }
  sub read
  {
    # read keys array
    read_keys($_[0]);
    # check all mandatory keywords
    while ( ($k, $t) = each %keywords ) {
      if ($t & 1) { 
        if (!defined $keys{$k}) { $err = "Mandatory keyword '$k' not found"; return $BAD; } 
      }
    }
    # convert short description to long if applicable
    if ((@{$keys{'ldesc'}} == 0) && $::config{'guesslongdesc'}) {
      @{$keys{'ldesc'}} = desc_convert( $keys{'desc'}->[0] );
    }
    # make inbound list
    if (!@in_dirs) {
      @in_dirs = @{$::config{'inbound'}};
      if (defined $::vars{'bad'}) { push @in_dirs, $::config{'badtics'}; }
    }
    # find file by crc
    undef $file;
    if ($::config{'crc_seek'} eq 'always') { $file = find_by_crc($keys{'crc'}, $keys{'size'}); }
    # find file by name
    if (!defined $file) {
      foreach $i (@in_dirs) { if (-f $i.'/'.$keys{'file'}) { $file = "$i/$keys{file}"; } }
      # try to find in all files by size/crc (crc check must be enabled)
      if (!defined $file && ($::config{'crc_seek'} eq 'missing') && !defined $::vars{'no_crc'}) { $file = find_by_crc($keys{'crc'}, $keys{'size'}); }
    }
    # give up
    if (!defined $file) {
      $err = "File $keys{file} not found"; return $BAD;
    }
    # find area by tag
    my $tag = uc $keys{'area'};
    undef %area;
    %area = %{$::areas{$tag}};
    # check CRC
    if ( !defined $::vars{'no_crc'} && defined %area && $area{'nocheck'} !~ /crc/i ) {
      #my $crc = `$::config{crc32} $file`;
      #if (hex $crc != hex $keys{'crc'}) { $err = "File $keys{file} CRC mismatch (in TIC=".sprintf("%x", $keys{'crc'})."; actual=".sprintf("%x", $crc).")"; return $BAD; }
      my $crc = ::file_crc32($file);
      if ($crc != hex $keys{'crc'}) { $err = "File $keys{file} CRC mismatch (in TIC=".lc($keys{'crc'})."; actual=".sprintf("%08x", $crc).")"; return $BAD; }
    }
    # check size, if present
    if ( !defined $::vars{'no_size'} && defined %area && $area{'nocheck'} !~ /size/i ) {
      if ( defined $keys{'size'} ) {
        if (-s $file != $keys{'size'}) { $err = "File $keys{file} size mismatch (in TIC=$keys{size}, actual=".(-s $file).")"; return $BAD; }
      }
    }
    # find sender node
    undef %node;
    my @afrom = ::fqfa2addr($keys{'from'});
    if ( !defined $afrom[0] ) { $err = "Bad from address $keys{from}"; return $BAD; }
    my $sfrom = ::addr4s(@afrom);
    if ( !defined $::nodes{$sfrom} ) { $err = "Unknown from-node $sfrom"; return $BAD; }
    my %node = %{$::nodes{$sfrom}};
    # check password for the link
    if ( !defined $::vars{'no_passwd'} && (uc $keys{'pw'} ne uc $node{'tic_pwd'}) ) { $err = "Wrong password for $sfrom: actual '$keys{pw}', expected '$node{tic_pwd}'"; return $BAD; }
    # check if the link can send us files or auto-create areas
    if ( defined %area ) {
      my $s = ::area_node(\%area, \@afrom);
      if ( !defined $s ) { $err = "Node $sfrom is not connected to area '$tag'"; return $BAD; }
      if ( $s =~ />/ && $s !~ /</ ) { $err = "Node $sfrom can't send us files in area '$tag'"; return $BAD; }
    } else {
      if ( !defined $node{'auto_add'} ) { $err = "Node $sfrom send us file in unknown area '$tag' and have no permissions to auto-add areas"; return $BAD; }
      if (!::create_area($tag, $node{'auto_add'}, $sfrom, \%keys)) { $err = "Undefined area $tag"; return $BAD; }
      %area = %{$::areas{$tag}};
    }
    # send ReturnReceipt message
    if ( defined $keys{'receiptrequest'} ) {
      @Tpl::from = @afrom; $Tpl::file = $file; $Tpl::keys = \%keys;
      $Tpl::aka = ::addr2s(@{$node{'use_aka'}});
      my $sysop = (defined $node{'sysop'}) ? $node{'sysop'} : 'Sysop';
      ::say 5, "> sending return receipt to $sfrom";
      my $s = do 'tpl/rrc.pl';
      ::send_mail( { from=>$::SYSOP, afrom=>$node{'use_aka'},
                     to=>$sysop, ato=>\@afrom,
                     subj=>'Return Receipt',
                     flags=>$Msg::PVT | $Msg::LOC | $Msg::RRC,
                     text=>$s
                   } ) if ($s ne '');
    }
    # everything seems ok
    return $OK;
  }
  # ------------------------------------------------------------------
  # Do filename to ms-dos 8.3 convertion, returns ($fnc_name, $orig_name)
  sub fnc 
  {
    my ($name, $ext) = $_[0] =~ /([^.]{1,8})[^.]*[.]?([^.]{0,3})/;
    $name =~ tr[0-9A-Za-z_^$~!#%&{}@`()'-][_]c;
    $ext  =~ tr[0-9A-Za-z_^$~!#%&{}@`()'-][_]c;
    $ext = '.'.$ext if $ext ne '';
    return ("$name$ext", $_[0]);
  }
  # ------------------------------------------------------------------
  # Do outgoing processing ($_[0] = flags, $_[1] - list of recipients)
  sub process
  {
    my $FLAG = (@_ > 0) ? $_[0] : 0;
    my $LIST = (@_ > 1) ? $_[1] : undef;
    # get file name
    #(my $fnc_name) = $file =~ m!([^/]*)$!;
    my ($fnc_name, $full_name) = fnc($keys{'file'});
    if ( !($FLAG & $NO_COPY) && defined $keys{'fullname'} ) { $full_name = $keys{'fullname'}; }
    # $name is the actual file name in local area
    my $name = ($::config{'filemode'} eq 'short') ? $fnc_name : $full_name;
    # ------------------------------------------------------
    # perform file filtering
    {
      my $act_policy = 1;
      foreach my $v (@{$area{'file'}}) {
        if ($name =~ /$v->[1]/i) { 
          ::say 5, "> rule $v->[0] /$v->[1]/ matched";
          $act_policy = ($v->[0] eq 'allow'); last; 
        }
        $act_policy = !($v->[0] eq 'allow');
      }
      if (!$act_policy) { ::say 2, "> file $name rejected"; }
    }
    # perform dupe-checking
    if ( !($FLAG & $NO_DUPE) && !defined $::vars{'no_dupe'} ) {
      if ( $area{'dcheck'} =~ /echo/i ) { $p1 = uc $keys{'area'}; } else { $p1 = '[^\t]*'; }
      if ( $area{'dcheck'} =~ /name/i ) { $p2 = uc $keys{'file'}; } else { $p2 = '[^\t]*'; }
      if ( $area{'dcheck'} =~ /crc/i )  { $p3 = lc $keys{'crc'}; } else { $p3 = '[^\t]*'; }
      if ( $::dupes =~ /^$p1\t$p2\t$p3\t.*$/m ) { ::say 2, "> duplicate TIC file"; return $DUPE; }
    }
    # create list of seen-by's
    my @seens = ();
    foreach my $l (@{$keys{'path'}}) {
      (my $s) = $l =~ /(\S+)/;
      my @a = ::fqfa2addr($s);
      if (defined $a[0]) { push @seens, \@a; }
    }
    foreach my $a (@seens) {
      if ( ::my_aka(@$a) ) { $err = "Loop in PATH found for local AKA ".::addr2s(@$a); return $BAD; }
    }
    foreach my $l (@{$keys{'seenby'}}) {
      my @a = ::fqfa2addr($l);
      if (defined $a[0]) { push @seens, \@a; }     
    }
    # do OnRecv action
    if ( defined $area{'onrecv'} ) {
      my $n = (defined $keys{'fullname'}) ? $keys{'fullname'} : $keys{'file'};
      my $D = join("\n", @{$keys{'desc'}});
      my $d = (@{$keys{'ldesc'}}) ? join("\n", @{$keys{'ldesc'}}) : $D;
      my $cmd = ::parse_symb($area{'onrecv'}, '$', {
                               'A'=>$tag, 'a'=>lc $tag,
                               'F'=>$file, 'f'=>(defined $area{'path'} ? $name : undef),
                               'N'=>$keys{'file'}, 'n'=>$n,
                               'P'=>$area{'path'}, 'p'=>$area{'path'},
                               'S'=>-s $file, 'C'=>lc $keys{'crc'},
                               'D'=>$D, 'd'=>$d, 'M'=>$keys{'magic'}
                             });
      ::say 5, "> executing command '$cmd'";
      $res = system $cmd;
      ::say 5, "> execution finished, exit code=".($res>>8);
    }
    # make sure file exists
    if ( !-f $file ) { ::say 2, "> file $file disappeared, skipping"; return $OK; }
    # ------------------------------------------------------
    # copy file to it's directory or pass-through directory
    # !important: for single mode should assure file with fnc-name!
    my $dir = ( defined $area{'path'} ) ? $area{'path'} : $::config{'queue'};
    # for 'link' and 'symlink' modes "$dir/$name" is outgoing file
    if ( $::config{'attachmode'} ne 'single' ) { ::copy $file, "$dir/$name" unless ($FLAG & $NO_COPY); }
    # for 'single' mode "$dir/$fnc_name" is outgoing file
    else { 
      ::copy $file, "$dir/$name" if defined $area{'path'} && !($FLAG & $NO_COPY);
      if ( !defined($area{'path'}) || ($name ne $fnc_name) ) {
        $dir = $::config{'queue'}; ::copy $file, "$dir/$fnc_name";
      }
    }
    if ( $::config{'filemode'} eq 'both' ) {
      symlink "$dir/$name", "$dir/$fnc_name";
    }
    if ( !($FLAG & $NO_COPY) && defined $area{'path'} ) {
      my $file = "$dir/$name";
      if (!chmod $area{'mode'}, $file) { ::say 3, "> warn: can't chmod $file to ".sprintf("%04o", $area{'mode'}); }
    }
    my $size = -s $file;
    # create file information record
    my %info = ();
    $info{'area'} = $keys{'area'};
    $info{'file'} = $keys{'file'};
    $info{'fnc_name'} = $fnc_name;
    $info{'full_name'} = $full_name;
    $info{'name'} = $name;
    $info{'size'} = $size;
    $info{'magic'} = $keys{'magic'} if defined $keys{'magic'};
    $info{'desc'} = join("\n", @{$keys{'desc'}});
    $info{'ldesc'} = (@{$keys{'ldesc'}}) ? join("\n", @{$keys{'ldesc'}}) : $info{'desc'};
    # add file to FILES.BBS
    if ( !($FLAG & $NO_FBBS) && defined($area{'path'}) && defined($::config{'filesbbs'}) ) {
      $Tpl::file = "$area{path}/$::config{filesbbs}";
      %Tpl::info = %info;
      do 'tpl/filesbbs.pl';
      undef $Tpl::file; undef %Tpl::info;
    }
    # do replaces if specified
    if ( !($FLAG & $NO_REPL) && defined $keys{'replaces'} && defined $area{'path'} ) {
      my @repl = glob "$area{path}/$keys{replaces}";
      foreach my $repl (@repl) {
        if ( ($::config{'filemode'} ne 'short') && (::basename($repl) eq $full_name) ) { next; }
        if ( ($::config{'filemode'} ne 'long') && (::basename($repl) eq $fnc_name) ) { next; }
        ::say 5, "> delete file $repl by REPLACES mask";
        $Tpl::info{'fnc_name'} = $repl; 
        if ( $::config{'filemode'} eq 'both' && -l $repl ) {
          my $repl_long = readlink $repl;
          unlink $repl; $repl = $repl_long;
        }
        unlink $repl;
        # remove file from FILES.BBS
        if ( !($FLAG & $NO_FBBS) && defined $::config{'filesbbs'} ) {
          $Tpl::file = "$area{path}/$::config{filesbbs}";
          $Tpl::info{'name'} = $Tpl::info{'full_name'} = $repl;
          $Tpl::del = 1;
          do 'tpl/filesbbs.pl';
          undef $Tpl::file; undef $Tpl::del;
        }
        undef %Tpl::info;
      }
    }
    # process MAGIC
    my $magic;
    if (defined $keys{'magic'}) {
    # perform magic filtering
    my $act_policy = 0;
    foreach my $v (@{$area{'magic'}}) {
      if ($keys{'magic'} =~ /$v->[1]/i) { 
        ::say 5, "> rule $v->[0] /$v->[1]/ matched";
        $act_policy = ($v->[0] eq 'allow'); last; 
      }
      $act_policy = !($v->[0] eq 'allow');
    }
    if (!$act_policy) { ::say 2, "> magic $keys{magic} rejected"; }
    elsif (defined $keys{'magic'} && defined $area{'path'}) {
      my $found = 0;
      for (my $i = 0; $i < @{$::config{'magicfile'}}; $i++) {
        my $m = $::config{'magicfile'}[$i];
        if (!open MF, '+<'.$m->[0]) { next; }
        binmode MF;
        my $modified = 0;
        my $s = join '', <MF>;
        # seek for a string
        my $ss = $m->[1]; $ss =~ s/[(][)]/$keys{'magic'}/i;
        my $s1 = $s; my $s0 = '';
        while ($s1 =~ /$ss/im) {
          my ($sr, $sf) = ($&, $1); $s1 = $';
          $s0 .= $`;
          if (!defined $1) { $s0 .= $sr; next; }
          my $target = ::rel2abs($area{'path'}.'/'.$name, $BASE_PATH);
          $sr =~ s!\Q$sf!$target!;
          $s = $s0.$sr.$s1; $s0 .= $sr; $modified = 1;
        }
        # save changes
        if ($modified) {
          seek MF, 0, 0; print MF $s; $found = 1;
        }
        # append magic
#        elsif ($i == @{$::config{'magicfile'}}-1) {
#          if ($s !~ /\n$/s) { print MF "\n"; }
#        }
        close MF;
      }
      if ($found) { $magic = $keys{'magic'}; }
    }
    # no action is required for 'replace' or 'update'
    if ($::config{'magicpass'} eq 'strip') { undef $magic; }
    elsif ($::config{'magicpass'} eq 'keep') { $magic = $keys{'magic'}; }
    }
    # ------------------------------------------------------
    # create list of recepients of current file
    my @to = ();
    my %my_seenby = ();
    foreach my $l (@{$area{'links'}}) {
      my $addr = $l->{'addr'};
      my $mode = $l->{'mode'};
      # check node record
      undef %node;
      my $sfrom = ::addr4s(@$addr);
      if ( !defined $::nodes{$sfrom} ) { next; }
      my %node = %{$::nodes{$sfrom}};
      # reject file sender
      my @sender = ::fqfa2addr($keys{'from'});
      if ( ::acmp($addr, \@sender) ) { next; }
      # check loops
      foreach my $s (@seens) {
        if ( ::acmp($addr, $s) == 1 ) { ::say 3, "> loop found for node ".::addr2s(@$addr)." in area $keys{area}"; next; }
      }
      # skip paused nodes
      if ( $node{'options'} =~ /pause/i ) { next; }
      # check if node can receive files
      if ( $mode =~ /</ && $mode !~ />/ ) { next; }
      # add node to @to-recv list and my aka to %my_seenby keys
      push @to, $addr;
        my $mya = $node{'use_aka'}; my $mys = 0;
        foreach my $s (@seens) { if ( ::acmp($mya, $s) == 1 ) { $mys = 1; last; } }
        $my_seenby{ ::addr4s @$mya } = 1 unless $mys;
    }
    # add to .tic missing values
    my @tic = ();
    push @tic, "Area ".uc $keys{'area'};
    push @tic, "Areadesc ".$keys{'areadesc'} if defined $keys{'areadesc'};
    push @tic, "Origin ".$keys{'origin'};
    my $pos_from = @tic; push @tic, "From ";
    #push @tic, "To ";
    push @tic, "Replaces ".$keys{'replaces'} if defined $keys{'replaces'};
    #push @tic, "Magic ";
    push @tic, "File $fnc_name";
    push @tic, "Fullname ".$full_name if ( $full_name ne $fnc_name );
    foreach (@{$keys{'desc'}}) { push @tic, "Desc ".$_; };
    foreach (@{$keys{'ldesc'}}) { push @tic, "LDesc ".$_; };
    push @tic, "Magic ".$magic if defined $magic;
    push @tic, "CRC ".$keys{'crc'};
    push @tic, "Size ".$size;
    push @tic, "Date ".$keys{'date'} if defined $keys{'date'};
    push @tic, "Release ".$keys{'release'} if defined $keys{'release'};
    push @tic, "Author ".$keys{'author'} if defined $keys{'author'};
    push @tic, "Source ".$keys{'source'} if defined $keys{'source'};
    foreach (@{$keys{'app'}}) { push @tic, "App ".$_; };
    push @tic, "Created by $::PRG_NAME $::PRG_VER $::PRG_COPYRIGHT";
    foreach (@{$keys{'via'}}) { push @tic, "Via ".$_; };
    if ( defined $keys{'created'} ) { 
      $_ = $keys{'created'}; s/^by //i; 
      $_ = "$keys{from} $_";
      push @tic, "Via ".$_; 
    }
    foreach (@{$keys{'path'}}) { push @tic, "Path ".$_; };
    my $pos_path = @tic; push @tic, "Path ";
    foreach (@{$keys{'seenby'}}) { push @tic, "Seenby ".$_; };
    foreach (keys %my_seenby) { push @tic, "Seenby ".$_; };
    foreach (@to) { push @tic, "Seenby ".(::addr2s @$_); };
    my $pos_pw = @tic; push @tic, "Pw ";
    push @tic, "PGP ".$keys{'pgp'} if defined $keys{'pgp'};
    push @tic, @ttic; # unknown keywords are passed as-is
    push @tic, "";    # just empty line
    # send file to all recepients
    foreach (@to) {
      my $sto = ::addr4s(@$_);
      my %node = %{$::nodes{$sto}};
      $tic[$pos_from] = "From ".(::addr4s @{$node{'use_aka'}});
      $tic[$pos_path] = "Path ".(::addr2s @{$node{'use_aka'}})." ".time()." ".gmtime()." UTC";
      $tic[$pos_pw] = "Pw ".$node{'tic_pwd'};
      # get filebox name for current node, if defined
      my $box = ::addr2box(\@$_); my $ndir;
      if (defined $box) { $ndir = $box; }
      else {
        # create temp directory in queue for node
        $ndir = $::config{'queue'}."/".(join '.', @$_[0..3]);
      }
      if ( ! -d $ndir ) { 
        if ( !mkdir($ndir, 0775) ) { ::say 3, "> can't create directory $ntic, skipping node $sto"; next; }
      }
      # get .tic name and write it
      my $t = time()%1000000;
      while ( -f "$ndir/ff".sprintf('%06d', $t).".tic" ) { $t--; }
      my $ntic = "$ndir/ff".sprintf('%06d', $t).".tic";
      ::say 5, "> writing .tic file for $sto to $ntic";
      open TIC, ">$ntic"; binmode TIC;
      print TIC join("\r\n", @tic);
      close TIC;
      # 
      if (defined $box) { ::say 5, "> writing file to filebox $ndir/$fnc_name"; }
      # create link to target file if need
      my $fdir = $dir;
      if ( $::config{'attachmode'} eq 'symlink' ) { symlink "$dir/$name", "$ndir/$fnc_name"; $fdir = '^'.$ndir; }
      elsif ( $::config{'attachmode'} eq 'link' ) { link "$dir/$name", "$ndir/$fnc_name"; $fdir = '^'.$ndir; }
      # in filebox mode, copy file to filebox for single-mode
      elsif (defined $box) { ::copy "$dir/$name", "$ndir/$fnc_name"; }
      # for BSO-mode
      if (!defined $box) {
        # if target system is busy (.bsy) - write to temp file instead of BSO
        my $bname = ::addr2bso(@$_); 
        { 
          (my $bdirname) = $bname =~ m!^(.*?)/[^/\\]*$!; 
          if ( ! -d $bdirname ) { mkdir $bdirname, 0775; }
        }
        my $bsy_name = '';
        my @bsy_glob = glob $::PAT_BSO.::pat_bso('bsy');
        #if ( -f $bname.'bsy' ) { $bname = $::config{'queue'}.'/delayed/'.join('.', @$_[0..3]).'.'; }
        if ( @bsy_glob ) { $bname = $::config{'queue'}.'/delayed/'.join('.', @$_[0..3]).'.'; }
          else { $bsy_name = $bname.'bsy'; open BSY, ">$bsy_name"; print BSY "$$\n"; close BSY; }
        # write .?lo-file
        my @bglob = glob $::PAT_BSO.::pat_bso( $node{'flag_att'}.'lo' );
        if ( @bglob ) { $bname = $bglob[0]; }
          else { $bname .= $node{'flag_att'}.'lo'; }
        ::say 5, "> writing .?lo file for $sto to $bname";
        open LO, ">>$bname";
        print LO "$fdir/$fnc_name\n^$ntic\n";
        close LO;
        if ( $bsy_name ne '' ) { unlink $bsy_name; }
      }
      # add file to node's send-to record
      push @{$::sendto{$sto}{uc $keys{'area'}}}, \%info;
    }
    # add .tic to dupes array
    $_ = sprintf("%s\t%s\t%08x\t%d\n", uc $keys{'area'}, $keys{'file'}, hex $keys{'crc'}, time);
    $::dupes .= $_;
    # add file info to daily report
    {
      #push @::report, \%info;
      my $s1 = pack 'u', $info{desc};  $s1 =~ tr!\n!!d;
      my $s2 = pack 'u', $info{ldesc}; $s2 =~ tr!\n!!d;
      $_ = time."\t".uc $keys{area}."\t".$keys{file}."\t".$name."\t".$size.
           "\t".sprintf("%08x", hex $keys{crc})."\t".$s1."\t".$s2."\t".$keys{magic}."\n";
      $::report .= $_;
    }
    # add parameters to traffic report file
    {
      my $trf = ::update_traffic($::traffic{ uc $keys{'area'} }, 1);
      $trf->{'cmf'}++; $trf->{'cwf'}++;
      $trf->{'cmk'} += $size; $trf->{'cwk'} += $size;
      #$trf->{'upd'} = time;
    }
    if ( !($FLAG & $NO_DEL) ) {
      # delete target file from inbound
      unlink $file;
      # delete pass-through file from queue in link mode (leave only links)
      if ( ($::config{'attachmode'} eq 'link') && !defined($area{'path'}) ) { unlink "$dir/$name"; }
    }
    # at last :)
    return $OK;
  }
}

1;
